package ch.std.jumpstartreactive;

import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.Test;
import org.reactivestreams.Subscriber;
import org.reactivestreams.Subscription;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.test.web.reactive.server.FluxExchangeResult;
import org.springframework.test.web.reactive.server.WebTestClient;

import ch.std.jumpstartreactive.beans.ReactiveCitiesBean;
import ch.std.jumpstartreactive.rest.ReactiveCityController;
import reactor.core.publisher.Flux;
import reactor.test.StepVerifier;

// see https://springframework.guru/spring-5-webclient/
// https://www.programcreek.com/java-api-examples/?api=reactor.test.StepVerifier

@SpringBootTest(webEnvironment = SpringBootTest.WebEnvironment.RANDOM_PORT)
public class JumpstartreactiveSubscribingApplicationTests {

	Logger logger = LoggerFactory.getLogger(ReactiveCityController.class);

	@Autowired
	private WebTestClient webTestClient;

	@Test
	public void contextLoads() {
	}

	@Test
	public void testGetCitiesTextEventStreamCheckAll() {
		FluxExchangeResult<String> fluxExchangeResult = webTestClient.get().uri("/cities")
				.accept(MediaType.TEXT_EVENT_STREAM).exchange().expectStatus().isOk().expectHeader()
				.contentTypeCompatibleWith(MediaType.TEXT_EVENT_STREAM).returnResult(String.class);

		Flux<String> cityFlux = fluxExchangeResult.getResponseBody();
		StepVerifier.create(cityFlux).recordWith(ArrayList::new)
				.expectNextCount(ReactiveCitiesBean.getSyncCities().size()).verifyComplete();
	}

	@Test
	public void testGetCitiesTextEventStreamFluxSubscribe() {
		FluxExchangeResult<String> fluxExchangeResult = webTestClient.get().uri("/cities")
				.accept(MediaType.TEXT_EVENT_STREAM).exchange().expectStatus().isOk().expectHeader()
				.contentTypeCompatibleWith(MediaType.TEXT_EVENT_STREAM).returnResult(String.class);

		Flux<String> cityFlux = fluxExchangeResult.getResponseBody();
		List<String> cityList = new ArrayList<>();
		cityFlux.log().subscribe(new Subscriber<String>() {
			@Override
			public void onSubscribe(Subscription s) {
				logger.info("onSubscribe");
				 s.request(Long.MAX_VALUE);
			}

			@Override
			public void onNext(String str) {
				logger.info("onNext, str = " + str);
				cityList.add(str);
			}

			@Override
			public void onError(Throwable t) {
			}

			@Override
			public void onComplete() {
				logger.info("onComplete");
			}
		});
	}

	@Test
	public void testGetCitiesTextEventStreamFluxLimitRate() throws InterruptedException {
		FluxExchangeResult<String> fluxExchangeResult = webTestClient.get().uri("/citieslimitrate")
				.accept(MediaType.TEXT_EVENT_STREAM).exchange().expectStatus().isOk().expectHeader()
				.contentTypeCompatibleWith(MediaType.TEXT_EVENT_STREAM).returnResult(String.class);

		Flux<String> cityFlux = fluxExchangeResult.getResponseBody();
		List<String> cityList = new ArrayList<>();
		cityFlux.log().subscribe(new Subscriber<String>() {
			@Override
			public void onSubscribe(Subscription s) {
				logger.info("onSubscribe");
				 s.request(Long.MAX_VALUE);
			}

			@Override
			public void onNext(String str) {
				logger.info("onNext, str = " + str);
				cityList.add(str);
			}

			@Override
			public void onError(Throwable t) {
			}

			@Override
			public void onComplete() {
				logger.info("onComplete");
			}
		});
		Thread.sleep(10000);
	}

	@Test
	public void testGetCitiesTextEventStreamFluxSubscribeBackpressing() {
		FluxExchangeResult<String> fluxExchangeResult = webTestClient.get().uri("/cities")
				.accept(MediaType.TEXT_EVENT_STREAM).exchange().expectStatus().isOk().expectHeader()
				.contentTypeCompatibleWith(MediaType.TEXT_EVENT_STREAM).returnResult(String.class);

		Flux<String> cityFlux = fluxExchangeResult.getResponseBody();
		List<String> cityList = new ArrayList<>();
		cityFlux.log().subscribe(new Subscriber<String>() {
			
			private Subscription s;
			private int nextIndex = 0;
			@Override
			public void onSubscribe(Subscription s) {
				this.s = s;
				logger.info("onSubscribe");
				 s.request(2);
			}

			@Override
			public void onNext(String str) {
				logger.info("onNext, s = " + str);
				cityList.add(str);
				if (nextIndex++ % 2 == 0) {
					try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
					}
					s.request(2);
				}
			}

			@Override
			public void onError(Throwable t) {
			}

			@Override
			public void onComplete() {
				logger.info("onComplete");
			}
		});
		logger.info("done");
	}

}
